"""Swagger"""

from .config import SwaggerConfig, SwaggerOauth2Config
from .repository import SwaggerRepository
from .controller import SwaggerController

__all__ = [
    "SwaggerController",
    "SwaggerRepository",
    "SwaggerConfig",
    "SwaggerOauth2Config",
]



"""Swagger Configuration"""

from typing import Callable, List, Optional


def _same_name(text: str) -> str:
    return text


class SwaggerOauth2Config:

    def __init__(
            self,
            client_id: str,
            realm: str,
            app_name: str,
            redirect_url: str
    ) -> None:
        """Swagger Oauth2 config

        Args:
            client_id (str): The client id
            realm (str): The realm
            app_name (str): The application name
            redirect_url (str): The redirect url
        """
        self.client_id = client_id
        self.realm = realm
        self.app_name = app_name
        self.redirect_url = redirect_url


class SwaggerConfig:

    def __init__(
        self,
        title: str = "The REST API",
        validator_url: Optional[str] = None,
        supported_submit_methods: Optional[List[str]] = None,
        display_operation_id: bool = False,
        display_request_duration: bool = False,
        doc_expansion: str = "list",
        oauth2: Optional[SwaggerOauth2Config] = None,
        serialize_key: Callable[[str], str] = _same_name,
        deserialize_key: Callable[[str], str] = _same_name
    ) -> None:
        """The Swagger UI Configuration

            Args:
                title (str, optional): The page title, defaults to "The REST API"
                validator_url (Optional[str], optional): The url of the validator,
                    defaults to None
                supported_submit_methods (Optional[List[str]], optional): The supported
                    submit methods, defaults to None
                display_operation_id (bool, optional): Whether operation ids should be
                    displayed, defaults to False
                display_request_duration (bool, optional): Whether the duration of the
                    request should be displayed, defaults to False
                doc_expansion (str, optional): How the methods should be displayed,
                    defaults to "list"
                oauth2 (Optional[SwaggerOauth2Config], optional): Optional oauth2
                    config, defaults to None
                serialize_key (Callable[[str], str]): The function to serialize keys
                deserialize_key (Callable[[str], str]): The function to deserialize keys

            """
        self.title = title
        self.validator_url = validator_url
        self.supported_submit_methods = supported_submit_methods
        self.display_operation_id = display_operation_id
        self.display_request_duration = display_request_duration
        self.doc_expansion = doc_expansion
        self.oauth2 = oauth2
        self.serialize_key = serialize_key
        self.deserialize_key = deserialize_key



"""Serialization"""

from cgi import parse_multipart
from functools import partial
import io
from typing import Any, Callable

from urllib.parse import parse_qs

from jetblack_serialization.config import SerializerConfig
from jetblack_serialization.types import Annotation
from jetblack_serialization.json import serialize, deserialize, from_json_value

from ..types import (
    MediaType,
    MediaTypeParams
)


def to_json(
        _media_type: MediaType,
        _params: MediaTypeParams,
        config: SerializerConfig,
        obj: Any,
        annotation: Any,
) -> str:
    """Convert the object to JSON

    Args:
        obj (Any): The object to convert

    Returns:
        str: The stringified object
    """
    return serialize(obj, annotation, config)


def from_json(
        _media_type: MediaType,
        _params: MediaTypeParams,
        config: SerializerConfig,
        text: str,
        annotation: Annotation
) -> Any:
    """Convert JSON to an object

    Args:
        text (str): The JSON string
        _media_type (bytes): The media type
        _params (Dict[bytes, bytes]): The params from content-type header
        annotation (str): The type annotation
        rename (Callable[[str], str]): A function to rename object keys.

    Returns:
        Any: The deserialized object.
    """
    return deserialize(text, annotation, config)


def from_query_string(
        _media_type: MediaType,
        _params: MediaTypeParams,
        _config: SerializerConfig,
        text: str,
        _annotation: Annotation
) -> Any:
    """Convert a query string to a dict

    Args:
        text (str): The query string
        _media_type (bytes): The media type from the content-type header.
        _params (Dict[bytes, bytes]): The params from the content-type header
        _annotation (str): The type annotation
        rename (Callable[[str], str]): A function to rename object keys.

    Returns:
        Any: The query string as a dict
    """
    return parse_qs(text)


def from_form_data(
        _media_type: MediaType,
        params: MediaTypeParams,
        _config: SerializerConfig,
        text: str,
        _annotation: Annotation
) -> Any:
    """Convert form data to a dict

    Args:
        text (str): The form data
        _media_type (bytes): The media type from the content-type header
        params (Dict[bytes, bytes]): The params from the content-type header.
        _annotation(str): The type annotation
        rename (Callable[[str], str]): A function to rename object keys.

    Raises:
        RuntimeError: If 'boundary' was not in the params

    Returns:
        Any: The form data as a dict.
    """
    if b'boundary' not in params:
        raise RuntimeError('Required "boundary" parameter missing')
    pdict = {
        name.decode(): value
        for name, value in params.items()
    }
    return parse_multipart(io.StringIO(text), pdict)


def json_arg_deserializer_factory(
        config: SerializerConfig,
) -> Callable[[str, Annotation], Any]:
    return partial(from_json_value, config)



"""Utility functions"""

import inspect
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence
)

from bareasgi.basic_router.path_definition import PathDefinition
import docstring_parser

from ..types import RestCallback

from .config import SwaggerConfig
from .parameters import make_swagger_parameters
from .responses import make_swagger_responses


def make_swagger_entry(
        method: str,
        path_definition: PathDefinition,
        callback: RestCallback,
        consumes: Sequence[bytes],
        produces: Sequence[bytes],
        collection_format: str,
        tags: Optional[List[str]],
        ok_status_code: int,
        ok_status_description: str,
        config: SwaggerConfig
) -> Dict[str, Any]:
    signature = inspect.signature(callback)
    docstring = docstring_parser.parse(inspect.getdoc(callback))
    params = make_swagger_parameters(
        method,
        consumes,
        path_definition,
        signature.parameters,
        docstring,
        collection_format,
        config
    )

    responses = make_swagger_responses(
        signature.return_annotation,
        docstring.returns if docstring else None,
        docstring.raises if docstring else None,
        ok_status_code,
        ok_status_description,
        collection_format,
        config
    )

    entry = {
        'parameters': params,
        'produces': [content_type.decode() for content_type in produces],
        'consumes': [accept.decode() for accept in consumes],
        'responses': responses
    }

    if docstring:
        if docstring.short_description:
            entry['summary'] = docstring.short_description
        if docstring.long_description:
            entry['description'] = docstring.long_description

    if tags:
        entry['tags'] = tags

    return entry



"""Constants"""

from bareasgi import HttpResponse, text_writer
from stringcase import camelcase, snakecase, pascalcase

from jetblack_serialization.config import SerializerConfig

from .types import (
    DictConsumes,
    DictProduces,
    DictSerializerConfig
)

from .serialization.json import (
    to_json,
    from_json,
    from_form_data,
    from_query_string,
    json_arg_deserializer_factory
)
from .serialization.xml import (
    from_xml,
    to_xml
)
from .swagger import SwaggerConfig

DEFAULT_SWAGGER_BASE_URL = "https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.4.0"
DEFAULT_TYPEFACE_URL = "https://fonts.googleapis.com/css?family=Roboto:300,400,500,700&display=swap"

DEFAULT_CONSUMES: DictConsumes = {
    b'application/json': from_json,
    b'application/x-www-form-urlencoded': from_query_string,
    b'multipart/form-data': from_form_data,
    b'application/xml': from_xml,
    b'*/*': from_json,
}

DEFAULT_PRODUCES: DictProduces = {
    b'application/json': to_json,
    b'application/xml': to_xml,
    b'*/*': to_json,
}

DEFAULT_COLLECTION_FORMAT = 'multi'

DEFAULT_NOT_FOUND_RESPONSE: HttpResponse = HttpResponse(
    404,
    [(b'content-type', b'text/plain')],
    text_writer('Not Found')
)

DEFAULT_JSON_SERIALIZER_CONFIG = SerializerConfig(camelcase, snakecase)
DEFAULT_XML_SERIALIZER_CONFIG = SerializerConfig(pascalcase, snakecase)

DEFAULT_SERIALIZER_CONFIG: DictSerializerConfig = {
    b'application/json': DEFAULT_JSON_SERIALIZER_CONFIG,
    b'*/*': DEFAULT_JSON_SERIALIZER_CONFIG,
    b'application/xml': DEFAULT_XML_SERIALIZER_CONFIG,
}

DEFAULT_ARG_DESERIALIZER_FACTORY = json_arg_deserializer_factory

DEFAULT_SWAGGER_CONFIG = SwaggerConfig(
    serialize_key=camelcase,
    deserialize_key=snakecase
)



"""Types"""

from typing import (
    Any,
    Awaitable,
    Callable,
    Dict
)

from jetblack_serialization.config import SerializerConfig
from jetblack_serialization.types import Annotation

MediaType = bytes
MediaTypeParams = Dict[bytes, bytes]

Deserializer = Callable[
    [
        MediaType,
        MediaTypeParams,
        SerializerConfig,
        str,
        Annotation
    ],
    Any
]
DictConsumes = Dict[bytes, Deserializer]

Serializer = Callable[
    [
        MediaType,
        MediaTypeParams,
        SerializerConfig,
        Any,
        Annotation
    ],
    str
]
DictProduces = Dict[bytes, Serializer]

DictSerializerConfig = Dict[bytes, SerializerConfig]

RestCallback = Callable[..., Awaitable[Any]]

ArgDeserializer = Callable[[str, Annotation], Any]

ArgDeserializerFactory = Callable[
    [SerializerConfig],
    Callable[[str, Annotation], Any]
]


class RestError(Exception):

    def __init__(self, status: int, message: str, *args) -> None:
        super().__init__(*args)
        self.status = status
        self.message = message



"""Argument builder"""

from inspect import Parameter, Signature
from typing import Any, Awaitable, Callable, Dict, List, Tuple

from jetblack_serialization.custom_annotations import (
    is_any_serialization_annotation
)

import jetblack_serialization.typing_inspect_ex as typing_inspect

from .types import ArgDeserializer


async def make_args(
        signature: Signature,
        matches: Dict[str, str],
        query: Dict[str, List[str]],
        body: Callable[[Any], Awaitable[Any]],
        arg_deserializer: ArgDeserializer
) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """Make args and kwargs for the given signature from the route matches,
    query args and body.

    Args:
        signature (Signature): The function signature
        matches (Dict[str, str]): The route matches
        query (Dict[str, Any]): A dictionary built from the query string
        body (Callable[[AsyncIterator[bytes], Any], Any]): Get the body
        arg_deserializer (ArgDeserializer): A deserializer for args

    Raises:
        KeyError: If a parameter was not found

    Returns:
        Tuple[Tuple[Any, ...], Dict[str, Any]]: A tuple for *args and **kwargs
    """

    kwargs: Dict[str, Any] = {}
    args: List[Any] = []

    for parameter in signature.parameters.values():
        if is_any_serialization_annotation(parameter.annotation):
            value: Any = await body(parameter.annotation)
        else:
            if parameter.name in matches:
                value = arg_deserializer(
                    matches[parameter.name],
                    parameter.annotation
                )
            elif parameter.name in query:
                if typing_inspect.is_list_type(parameter.annotation):
                    element_type, *_rest = typing_inspect.get_args(
                        parameter.annotation
                    )
                    value = [
                        arg_deserializer(
                            item,
                            element_type
                        )
                        for item in query[parameter.name]
                    ]
                else:
                    value = arg_deserializer(
                        query[parameter.name][0],
                        parameter.annotation
                    )

            elif typing_inspect.is_optional_type(parameter.annotation):
                value = None
            else:
                raise KeyError(parameter.name)

        if (
                parameter.kind == Parameter.POSITIONAL_ONLY
                or parameter.kind == Parameter.POSITIONAL_OR_KEYWORD
        ):
            args.append(value)
        else:
            kwargs[parameter.name] = value

    bound_args = signature.bind(*args, **kwargs)
    bound_args.apply_defaults()

    return bound_args.args, bound_args.kwargs



"""Utility functions"""

from typing import (
    Generic,
    TypeVar
)

T = TypeVar('T')


class NullIter(Generic[T]):
    """An iterator containing no items"""

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        raise StopAsyncIteration



"""Utility functions"""

import inspect
from typing import (
    Any,
    Dict,
    List,
    Optional
)

from docstring_parser import DocstringReturns, DocstringRaises

from .config import SwaggerConfig
from .errors import gather_error_responses
from .properties import get_property


def make_swagger_responses(
        return_annotation: Any,
        docstring_returns: Optional[DocstringReturns],
        docstring_raises: Optional[List[DocstringRaises]],
        ok_status_code: int,
        ok_status_description: str,
        collection_format: str,
        config: SwaggerConfig
) -> Dict[int, Dict[str, Any]]:
    ok_response: Dict[str, Any] = {
        'description': ok_status_description
    }

    if return_annotation is not None:
        ok_response['schema'] = get_property(
            return_annotation,
            None,
            docstring_returns.description if docstring_returns else None,
            inspect.Parameter.empty,
            collection_format,
            config
        )

    responses: Dict[int, Dict[str, Any]] = {
        ok_status_code: ok_response
    }
    if docstring_raises:
        error_responses = gather_error_responses(docstring_raises)
        responses.update(error_responses)

    return responses



"""Utility functions"""

from bareasgi.basic_router.path_definition import PathDefinition


def make_swagger_path(path_definition: PathDefinition) -> str:
    """Make a path compatible with swagger"""
    swagger_path = '/' + '/'.join(
        '{' + segment.name + '}' if segment.is_variable else segment.name
        for segment in path_definition.segments
    )
    if path_definition.ends_with_slash:
        swagger_path += '/'
    return swagger_path



"""Type Info"""

from datetime import datetime, timedelta
from decimal import Decimal
import inspect
from typing import (
    Any,
    Dict,
    Optional
)

import docstring_parser
from docstring_parser import Docstring

from jetblack_serialization.custom_annotations import (
    get_default_annotation,
    is_any_default_annotation
)
from jetblack_serialization.types import Annotation
import jetblack_serialization.typing_inspect_ex as typing_inspect

from .config import SwaggerConfig
from .utils import find_docstring_param


def get_property(
        annotation: Any,
        name: Optional[str],
        description: Optional[str],
        default: Any,
        collection_format: str,
        config: SwaggerConfig
) -> Dict[str, Any]:
    """Get a swagger property

    Args:
        annotation (Any): The type annotation
        name (Optional[str]): An optional property name
        description (Optional[str]): An optional property description
        default (Any): An optional default where inspect.Parameter.empty indicates no default
        collection_format (str): The swagger collection format

    Raises:
        TypeError: If the property type is not handled.

    Returns:
        Dict[str, Any]: The swagger property.
    """
    if typing_inspect.is_annotated_type(annotation):
        return get_property(
            typing_inspect.get_origin(annotation),
            name,
            description,
            default,
            collection_format,
            config
        )

    if typing_inspect.is_optional_type(annotation):
        optional_type = typing_inspect.get_optional_type(annotation)
        return get_property(
            optional_type,
            name,
            description,
            default,
            collection_format,
            config
        )

    prop: Dict[str, Any] = {}

    if name:
        prop['name'] = name

    if description:
        prop['description'] = description

    if default != inspect.Parameter.empty:
        prop['default'] = default

    if annotation is str:
        prop['type'] = 'string'
    elif annotation is bool:
        prop['type'] = 'boolean'
    elif annotation is int:
        prop['type'] = 'integer'
    elif annotation is float:
        prop['type'] = 'number'
    elif annotation is Decimal:
        prop['type'] = 'number'
    elif annotation is datetime:
        prop['type'] = 'string'
        prop['format'] = 'date-time'
    elif annotation is timedelta:
        # Note: Swagger has no support for durations. I made up the format.
        prop['type'] = 'string'
        prop['format'] = 'duration'
    elif typing_inspect.is_list_type(annotation):
        contained_type, *_rest = typing_inspect.get_args(annotation)
        prop['type'] = 'array'
        prop['collectionFormat'] = collection_format
        prop['items'] = get_property(
            contained_type,
            None,
            None,
            default,
            collection_format,
            config
        )
    elif typing_inspect.is_dict_type(annotation):
        prop['type'] = 'object'
    elif typing_inspect.is_typed_dict_type(annotation):
        prop['type'] = 'object'
        prop['properties'] = get_properties(
            annotation,
            docstring_parser.parse(inspect.getdoc(annotation)),
            collection_format,
            config
        )
    else:
        raise TypeError('Unhandled type annotation')

    return prop


def _get_default(annotation, member_annotation, name) -> Any:
    if is_any_default_annotation(member_annotation):
        _, default_value = get_default_annotation(member_annotation)
        return default_value.value

    return getattr(annotation, name, inspect.Parameter.empty)


def get_properties(
        annotation,
        docstring: Docstring,
        collection_format: str,
        config: SwaggerConfig
) -> Dict[str, Any]:
    """Get the properties of a TypedDict

    Args:
        annotations (Dict[str, Annotation]): The member
            annotations
        docstring (Docstring): The docstring
        collection_format (str): The collection format

    Returns:
        Dict[str, Any]: The swagger properties.
    """
    annotations: Dict[str, Annotation] = typing_inspect.typed_dict_keys(
        annotation
    )
    properties: Dict[str, Any] = {}
    for name, member_annotation in annotations.items():
        camelcase_name = config.serialize_key(name)
        docstring_param = find_docstring_param(name, docstring)
        description = docstring_param.description if docstring_param else None
        default = _get_default(annotation, member_annotation, name)

        properties[camelcase_name] = get_property(
            member_annotation,
            camelcase_name,
            description,
            default,
            collection_format,
            config
        )

    return properties



"""Swagger Utility functions"""

from typing import Optional

from docstring_parser import Docstring, DocstringParam


def find_docstring_param(
        name: str,
        docstring: Docstring
) -> Optional[DocstringParam]:
    """Find the docstring param for the given parameter name.

    Args:
        name (str): The parameter name
        docstring (Docstring): The docstring

    Returns:
        Optional[DocstringParam]: The docstring param or None if not found.
    """
    for param in docstring.params:
        if param.arg_name == name:
            return param
    return None



"""Helpers"""

from typing import Any

from jetblack_serialization.config import SerializerConfig
from jetblack_serialization.types import Annotation
from jetblack_serialization.xml import serialize, deserialize

from ..types import (
    MediaType,
    MediaTypeParams
)


def from_xml(
        _media_type: MediaType,
        _params: MediaTypeParams,
        config: SerializerConfig,
        text: str,
        annotation: Annotation
) -> Any:
    return deserialize(text, annotation, config)


def to_xml(
        _media_type: MediaType,
        _params: MediaTypeParams,
        config: SerializerConfig,
        obj: Any,
        annotation: Any,
) -> str:
    return serialize(obj, annotation, config)



"""Parameters"""

from inspect import Parameter
from typing import (
    AbstractSet,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set
)

from bareasgi.basic_router.path_definition import PathDefinition
from docstring_parser import Docstring, DocstringParam
import jetblack_serialization.typing_inspect_ex as typing_inspect
from jetblack_serialization.custom_annotations import (
    is_any_serialization_annotation
)

from .config import SwaggerConfig
from .properties import get_property
from .utils import find_docstring_param


def _make_swagger_parameter(
        source: str,
        param: Parameter,
        collection_format: str,
        docstring_param: Optional[DocstringParam],
        config: SwaggerConfig
) -> Dict[str, Any]:
    is_required = param.default is Parameter.empty

    prop = get_property(
        param.annotation,
        config.serialize_key(param.name),
        docstring_param.description if docstring_param else None,
        param.default,
        collection_format,
        config
    )

    if source != 'body':
        prop['in'] = source
        prop['required'] = is_required

    return prop


def _make_swagger_parameters_inline(
        source: str,
        parameters: Mapping[str, Parameter],
        path_variables: AbstractSet[str],
        docstring: Docstring,
        collection_format: str,
        config: SwaggerConfig
) -> List[Dict[str, Any]]:
    """Make inline paramters for query or form"""
    props: List[Dict[str, Any]] = []
    for parameter in parameters.values():
        if parameter.name in path_variables:
            continue
        docstring_param = find_docstring_param(parameter.name, docstring)
        props.append(
            _make_swagger_parameter(
                source,
                parameter,
                collection_format,
                docstring_param,
                config
            )
        )
    return props


def make_swagger_parameters(
        method: str,
        consumes: Sequence[bytes],
        path_definition: PathDefinition,
        parameters: Mapping[str, Parameter],
        docstring: Docstring,
        collection_format: str,
        config: SwaggerConfig
) -> List[Dict[str, Any]]:
    """Make the swagger parameters"""

    available_parameters = {
        name: parameter
        for name, parameter in parameters.items()
    }

    # Path parameters
    props: List[Dict[str, Any]] = []
    path_variables: Set[str] = set()
    for segment in path_definition.segments:
        if segment.is_variable:
            path_variable = config.deserialize_key(segment.name)
            path_variables.add(path_variable)
            parameter = available_parameters.pop(path_variable)
            docstring_param = find_docstring_param(parameter.name, docstring)
            prop = _make_swagger_parameter(
                'path',
                parameter,
                collection_format,
                docstring_param,
                config
            )
            props.append(prop)

    if method == 'GET':
        # Query parameters
        props.extend(
            _make_swagger_parameters_inline(
                'query',
                available_parameters,
                path_variables,
                docstring,
                collection_format,
                config
            )
        )
    elif (
            b'application/x-www-form-urlencoded' in consumes or
            b'multipart/form-data' in consumes
    ):
        # Form body parameters
        props.extend(
            _make_swagger_parameters_inline(
                'formData',
                available_parameters,
                path_variables,
                docstring,
                collection_format,
                config
            )
        )
    else:
        # Fall back to b'application/json'.
        for parameter in available_parameters.values():
            docstring_param = find_docstring_param(parameter.name, docstring)
            if is_any_serialization_annotation(parameter.annotation):
                body_type = typing_inspect.get_origin(parameter.annotation)
                schema = get_property(
                    body_type,
                    config.serialize_key(parameter.name),
                    None,
                    Parameter.empty,
                    collection_format,
                    config
                )
                props.append({
                    'in': 'body',
                    'name': 'schema',
                    'description': 'The body schema',
                    'schema': schema
                })
            else:
                props.append(
                    _make_swagger_parameter(
                        'query',
                        parameter,
                        collection_format,
                        docstring_param,
                        config
                    )
                )

    return props



"""A swagger controller"""

import logging
import json
from typing import Optional

from bareasgi import HttpRequest, HttpResponse, text_writer
from bareasgi.basic_router.http_router import BasicHttpRouter
from bareasgi_jinja2 import Jinja2TemplateProvider
from bareutils import response_code

from .repository import SwaggerRepository
from .config import SwaggerConfig

LOGGER = logging.getLogger(__name__)


class SwaggerController:
    """A swagger controller"""

    def __init__(
            self,
            title: str,
            base_path: str,
            swagger_base_url: str,
            typeface_url: str,
            config: Optional[SwaggerConfig],
            repo: SwaggerRepository
    ) -> None:
        self.title = title
        self.base_path = base_path
        self.swagger_base_url = swagger_base_url
        self.typeface_url = typeface_url
        self.config = config or SwaggerConfig()

        self.repo = repo

    def add_routes(self, router: BasicHttpRouter):
        """Add the swagger routes

        Args:
            router (BasicHttpRouter): The router
        """
        router.add(
            {'GET'},
            self.base_path + '/swagger.json',
            self._swagger_json
        )
        router.add(
            {'GET'},
            self.base_path + '/swagger',
            self._swagger_ui
        )

    async def _swagger_json(self, _request: HttpRequest) -> HttpResponse:
        body = text_writer(json.dumps(self.repo.definition))
        headers = [(b'content-type', b'application/json')]
        return HttpResponse(response_code.OK, headers, body)

    async def _swagger_ui(self, request: HttpRequest) -> HttpResponse:
        return await Jinja2TemplateProvider.apply(
            request,
            'swagger.html',
            {
                "title": self.title,
                "specs_url": self.base_path + "/swagger.json",
                'swagger_base_url': self.swagger_base_url,
                'typeface_url': self.typeface_url,
                "config": self.config
            }
        )



"""Swagger errors"""

from typing import Any, Dict, List

from docstring_parser import DocstringRaises


def gather_error_responses(docstring_raises: List[DocstringRaises]) -> Dict[int, Any]:
    """Gather error responses

    Looks for exceptions of type `HTTPError` with a description starting with
    the error code: e.g. `"404, when a book is not found"`

    Args:
        docstring_raises (List[DocstringRaises]): The raises from the docstring

    Returns:
        Dict[int, Any]: The error response schema.
    """
    responses: Dict[int, Any] = {}
    for raises in docstring_raises:
        if raises.type_name != 'HTTPError':
            continue
        first, sep, rest = raises.description.partition(',')
        if not sep:
            continue
        try:
            error_code = int(first.strip())
            description = rest.strip()
            responses[error_code] = {
                'description': description
            }
        except:  # pylint: disable=bare-except
            continue
    return responses



"""A router for REST APIs

Attributes:
    DEFAULT_SWAGGER_BASE_URL (str): The default swagger CDN url. The currently
        supported version is 3.4.0
    DEFAULT_TYPEFACE_URL (str): The typeface url to use.
"""

import inspect
import logging
from typing import (
    AbstractSet,
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    cast
)
from urllib.parse import parse_qs

from bareasgi import HttpRequest, HttpResponse, text_reader, text_writer
from bareasgi.basic_router.http_router import BasicHttpRouter, PathDefinition
from bareutils import header, response_code
from jetblack_serialization.config import SerializerConfig

from .arg_builder import make_args
from .swagger import SwaggerRepository, SwaggerConfig, SwaggerController
from .constants import (
    DEFAULT_SWAGGER_BASE_URL,
    DEFAULT_TYPEFACE_URL,
    DEFAULT_CONSUMES,
    DEFAULT_PRODUCES,
    DEFAULT_COLLECTION_FORMAT,
    DEFAULT_NOT_FOUND_RESPONSE,
    DEFAULT_SERIALIZER_CONFIG,
    DEFAULT_JSON_SERIALIZER_CONFIG,
    DEFAULT_ARG_DESERIALIZER_FACTORY,
    DEFAULT_SWAGGER_CONFIG
)
from .types import (
    Deserializer,
    DictConsumes,
    DictProduces,
    DictSerializerConfig,
    RestCallback,
    ArgDeserializerFactory,
    RestError,
    Serializer
)

LOGGER = logging.getLogger(__name__)


def _rename_path_definition(
        path_definition: PathDefinition,
        config: SerializerConfig
) -> PathDefinition:
    for segment in path_definition.segments:
        if segment.is_variable:
            segment.name = config.serialize_key(
                segment.name
            ) if config.serialize_key else segment.name
    return path_definition


class RestHttpRouter(BasicHttpRouter):
    """A REST router"""

    def __init__(
            self,
            title: str,
            version: str,
            *,
            not_found_response: Optional[HttpResponse] = None,
            description: Optional[str] = None,
            base_path: str = '',
            consumes: Optional[DictConsumes] = None,
            produces: Optional[DictProduces] = None,
            tags: Optional[List[Mapping[str, Any]]] = None,
            swagger_base_url: str = DEFAULT_SWAGGER_BASE_URL,
            typeface_url: str = DEFAULT_TYPEFACE_URL,
            config: SwaggerConfig = DEFAULT_SWAGGER_CONFIG,
            serializer_configs: Optional[DictSerializerConfig] = None,
            arg_serializer_config: SerializerConfig = DEFAULT_JSON_SERIALIZER_CONFIG,
            arg_deserializer_factory: ArgDeserializerFactory = DEFAULT_ARG_DESERIALIZER_FACTORY
    ) -> None:
        """Initialise the REST router

        Here is an example of how to use the router.

        ```python
        from bareasgi import Application
        from bareasgi_rest import RestHttpRouter, add_swagger_ui

        router = RestHttpRouter(
            title="Books",
            version="1",
            description="A book api",
            base_path='/api/1',
            tags=[
                {
                    'name': 'Books',
                    'description': 'The book store API'
                }
            ]
        )
        app = Application(http_router=router)
        add_swagger_ui(app)
        ```

        Args:
            title (str): The title of the swagger documentation.
            version (str): The version of the exposed API.
            not_found_response (Optional[HttpResponse], optional): The response
                sent when a route is not found. Defaults to None.
            description (Optional[str], optional): The API description. Defaults
                to None.
            base_path (str, optional): The base path of the API. Defaults to ''.
            consumes (Optional[DictConsumes], optional): A map of media types
                and deserializers. Defaults to DEFAULT_CONSUMES.
            produces (Optional[DictProduces], optional): A map of media types
                and serializers. Defaults to DEFAULT_PRODUCES.
            tags (Optional[List[Mapping[str, Any]]], optional): The available
                tags. Defaults to None.
            swagger_base_url (Optional[str], optional): The base url for the
                swagger CDN. Defaults to DEFAULT_SWAGGER_BASE_URL.
            typeface_url (Optional[str], optional): The base url for the
                typeface. Defaults to DEFAULT_TYPEFACE_URL.
            config (Optional[SwaggerConfig], optional): The swagger
                configuration. Defaults to None.
            serializer_configs (Optional[DictSerializerConfig], optional): The
                serializer configuration for content. Defaults to
                DEFAULT_SERIALIZER_CONFIG.
            arg_serializer_config (SerializerConfig, optional): The serializer
                configuration for arguments. Defaults to DEFAULT_JSON_SERIALIZER_CONFIG.
            arg_deserializer_factory (ArgDeserializerFactory, optional): The
                deserializer configuration for arguments. Defaults to
                DEFAULT_ARG_DESERIALIZER_FACTORY.
        """
        super().__init__(not_found_response or DEFAULT_NOT_FOUND_RESPONSE)
        self.consumes = consumes or DEFAULT_CONSUMES
        self.produces = produces or DEFAULT_PRODUCES
        self.base_path = base_path

        self.accepts: Dict[str, Dict[PathDefinition, bytes]] = {}
        self.collection_formats: Dict[str, Dict[PathDefinition, str]] = {}

        self.serializer_configs = serializer_configs or DEFAULT_SERIALIZER_CONFIG
        self.arg_serializer_config = arg_serializer_config
        self.arg_deserializer_factory = arg_deserializer_factory

        self.swagger_repo = SwaggerRepository(
            title,
            version,
            description,
            base_path,
            [name.decode() for name in self.consumes.keys()],
            [name.decode() for name in self.produces.keys()],
            tags,
            config
        )
        self.swagger_controller = SwaggerController(
            title,
            base_path,
            swagger_base_url,
            typeface_url,
            config,
            self.swagger_repo
        )
        self.swagger_controller.add_routes(self)

    def add_rest(
            self,
            methods: AbstractSet[str],
            path: str,
            callback: RestCallback,
            *,
            consumes: Optional[Sequence[bytes]] = None,
            produces: Optional[Sequence[bytes]] = None,
            collection_format: str = DEFAULT_COLLECTION_FORMAT,
            tags: Optional[List[str]] = None,
            status_code: int = response_code.OK,
            status_description: str = 'OK',
            serializer_config: Optional[DictSerializerConfig] = None,
            arg_serializer_config: Optional[SerializerConfig] = None,
            arg_deserializer_factory: Optional[ArgDeserializerFactory] = None
    ) -> None:
        """Register a callback to a method and path

        Args:
            methods (AbstractSet[str]): The set of methods
            path (str): The path
            callback (RestCallback): The callback
            produces (List[bytes], optional): The accept media type. Defaults to
                None.
            consumes (List[bytes], optional): The content media type. Defaults
                to None.
            collection_format (str, optional): The format of repeated values.
                Defaults to DEFAULT_COLLECTION_FORMAT.
            tags (Optional[List[str]], optional): A list of tags. Defaults to
                None.
            status_code (int, optional): The ok status code. Defaults to 200.
            status_description (str, optional): The ok status message. Defaults
                to 'OK'.
            serializer_config (Optional[DictSerializerConfig], optional): The
                serializer configuration for content. Defaults to None.
            arg_serializer_config (Optional[SerializerConfig], optional): The
                serializer configuration for arguments. Defaults to None.
            arg_deserializer_factory (Optional[ArgDeserializerFactory], optional): The
                deserializer configuration for arguments. Defaults to None.
        """
        LOGGER.debug('Adding route for %s on "%s"', methods, path)

        if produces is None:
            produces = list(self.produces.keys())
        if consumes is None:
            consumes = list(self.consumes.keys())

        for method in methods:
            self._add_method(
                method,
                path,
                callback,
                produces,
                status_code,
                serializer_config,
                arg_serializer_config,
                arg_deserializer_factory
            )
            self.swagger_repo.add(
                method,
                _rename_path_definition(
                    PathDefinition(path),
                    DEFAULT_JSON_SERIALIZER_CONFIG
                ),
                callback,
                consumes,
                produces,
                collection_format,
                tags,
                status_code,
                status_description
            )

    def _add_method(
            self,
            method: str,
            path: str,
            callback: RestCallback,
            produces: Sequence[bytes],
            status_code: int,
            serializer_configs: Optional[DictSerializerConfig],
            arg_serializer_config: Optional[SerializerConfig],
            arg_deserializer_factory: Optional[ArgDeserializerFactory]
    ) -> None:
        signature = inspect.signature(callback)
        path_definition = _rename_path_definition(
            PathDefinition(self.base_path + path),
            DEFAULT_JSON_SERIALIZER_CONFIG
        )

        arg_deserializer = (
            arg_deserializer_factory or self.arg_deserializer_factory
        )(
            arg_serializer_config or self.arg_serializer_config
        )

        async def rest_callback(request: HttpRequest) -> HttpResponse:

            route_args: Dict[str, str] = {
                self.arg_serializer_config.deserialize_key(name): value
                for name, value in request.matches.items()
            }
            query_string = request.scope['query_string'].decode()
            query_args: Dict[str, List[str]] = {
                self.arg_serializer_config.deserialize_key(name): values
                for name, values in parse_qs(query_string).items()
            }
            body_reader = self._get_body_reader(request)

            try:
                args, kwargs = await make_args(
                    signature,
                    route_args,
                    query_args,
                    body_reader,
                    arg_deserializer
                )
            except BaseException as error:  # pylint: disable=broad-except
                return HttpResponse(
                    response_code.BAD_REQUEST,
                    [(b'content-type', b'text/plain')],
                    text_writer("Failed to make args:" + ". ".join(error.args))
                )

            try:
                body = await callback(*args, **kwargs)
            except RestError as error:
                return HttpResponse(
                    error.status,
                    [(b'content-type', b'text/plain')],
                    text_writer(error.message)
                )
            except BaseException as error:  # pylint: disable=broad-except
                return HttpResponse(
                    response_code.INTERNAL_SERVER_ERROR,
                    [(b'content-type', b'text/plain')],
                    text_writer(str(error))
                )

            accept = header.accept(request.scope['headers'])
            writer = self._make_writer(
                body,
                accept,
                signature.return_annotation,
                serializer_configs or self.serializer_configs
            )
            if not accept:
                content_type = produces[0]
            else:
                for content_type in produces:
                    if content_type in accept:
                        break
                else:
                    if b'*/*' in accept:
                        # Prefer the first content type that is supported.
                        content_type = produces[0]
                    else:
                        return HttpResponse(
                            response_code.UNSUPPORTED_MEDIA_TYPE,
                            [(b'content-type', b'text/plain')],
                            text_writer('Unsupported media type')
                        )

            headers = [
                (b'content-type', content_type)
            ]

            return HttpResponse(status_code, headers, writer)

        self.add_route(method, path_definition, rest_callback)

    def _make_writer(
            self,
            data: Optional[Any],
            accept: Optional[Mapping[bytes, Mapping[bytes, Any]]],
            return_annotation: Any,
            serializer_configs: DictSerializerConfig
    ) -> Optional[AsyncIterable[bytes]]:
        if data is None:
            # No need for a writer if there is no data.
            return None

        # Prefer the media types in the order they are defined.
        media_type: Optional[bytes] = None
        serializer: Optional[Serializer] = None
        if accept:
            for media_type, serializer in self.produces.items():
                if media_type in accept:
                    break
        else:
            # If no accept choose the first from produces.
            media_type, serializer = next(iter(self.produces.items()))

        if media_type is None or serializer is None:
            raise ValueError(f'No handler for media types: {accept.keys()}')

        serializer_config = serializer_configs[media_type]

        text = serializer(
            media_type,
            {},
            serializer_config,
            data,
            return_annotation
        )
        return text_writer(text)

    def _get_body_reader(
            self,
            request: HttpRequest
    ) -> Callable[[Any], Awaitable[Any]]:
        if request.scope['method'] in {'GET'}:
            deserializer: Optional[Deserializer] = None
            serializer_config: SerializerConfig = DEFAULT_JSON_SERIALIZER_CONFIG
        else:
            media_type, params = header.content_type(
                request.scope['headers']
            ) or (b'application/json', cast(Dict[bytes, Any], {}))
            deserializer = self.consumes[media_type]
            serializer_config = self.serializer_configs[media_type]

        async def body_reader(annotation: Any) -> Any:
            if deserializer is None:
                raise RuntimeError('No deserializer')
            text = await text_reader(request.body)
            return deserializer(
                media_type,
                cast(Dict[bytes, Any], params),
                serializer_config,
                text,
                annotation
            )

        return body_reader



"""A Swagger Repository"""


from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence
)

from bareasgi.basic_router.http_router import PathDefinition

from ..types import RestCallback

from .config import SwaggerConfig
from .entry import make_swagger_entry
from .paths import make_swagger_path


class SwaggerRepository:
    """A swagger repository"""

    def __init__(
            self,
            title: str,
            version: str,
            description: Optional[str],
            base_path: str,
            consumes: Optional[List[str]],
            produces: Optional[List[str]],
            tags: Optional[List[Mapping[str, Any]]],
            config: SwaggerConfig
    ) -> None:
        self.config = config

        self.definition: Dict[str, Any] = {
            'swagger': '2.0',
            'basePath': base_path,
            'info': {
                'title': title,
                'version': version,
                'description': description
            },
            'produces': produces or [],
            'consumes': consumes or [],
            "paths": {},
        }
        if tags:
            self.definition['tags'] = tags

    def add(
            self,
            method: str,
            path_definition: PathDefinition,
            callback: RestCallback,
            consumes: Sequence[bytes],
            produces: Sequence[bytes],
            collection_format: str,
            tags: Optional[List[str]],
            status_code: int,
            status_description: str
    ):
        """Add a swagger entry

        Args:
            method (str): The HTTP method
            path_definition (PathDefinition): The router path
            callback (RestCallback): The callback
            consumes (Sequence[bytes]): The accept header
            produces (Sequence[bytes]): The content type
            collection_format (str): The collection format
            tags (Optional[List[str]]): Optional tags
            status_code (int): The ok status code
            status_description (str): The ok status description
        """

        entry = make_swagger_entry(
            method,
            path_definition,
            callback,
            consumes,
            produces,
            collection_format,
            tags,
            status_code,
            status_description,
            self.config
        )

        swagger_path = make_swagger_path(path_definition)

        paths: Dict[str, Any] = self.definition['paths']
        current_path: Dict[str, Any] = paths.setdefault(swagger_path, {})
        current_path[method.lower()] = entry
